//--------------------------------------------------------------------------------------
// WebServices.cpp
//
// Demonstrates how to get an Authorization token from Xbox and use it with an Xbox
// Web Service or a title's custom web services. The resulting response data is parsed
// and displayed.
//
// Advanced Technology Group (ATG)
// Copyright (C) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#include "pch.h"
#include "BasicTimer.h"

using namespace Microsoft;
using namespace Windows::Xbox::System;
using namespace XboxSampleFramework;

// TODO: Replace with your server's host name
#define CUSTOMURI_REQUIRE_XSTS      FALSE
#define CUSTOMURI_HOST              L"YOURSERVER"
#define CUSTOMURI_REQUEST_URL_1     L"https://" CUSTOMURI_HOST L"/restservice.svc/getclaims"
#define CUSTOMURI_REQUEST_URL_2     L"https://" CUSTOMURI_HOST L"/restservice.svc/messageoftheday"
#define CUSTOMURI_REQUEST_BODY      L"{\"message\":\"Hello\"}"

//  These API's are intended for use with the Delegated Auth sample
//  To switch this app, set the below definition to true
#define ENABLE_DELEGATED_URLS       FALSE
#define CONSUMABLE_PRODUCT_ID       L"SET_YOUR_PRODUCT_ID"
#define DELEGATED_REQUEST_INVENTORY L"https://" CUSTOMURI_HOST L"/restservice.svc/Inventory"
#define DELEGATED_REQUEST_CONSUME   L"https://" CUSTOMURI_HOST L"/restservice.svc/Consume"
#define DELEGATED_REQUEST_PURCHASE  L"https://" CUSTOMURI_HOST L"/restservice.svc/Purchase"
#define DELEGATED_REQUEST_FRIENDS   L"https://" CUSTOMURI_HOST L"/restservice.svc/Friends"

// On screen help
const LPCWSTR SAMPLE_TITLE_TEXT       = L"Web Service Communication";
const LPCWSTR SAMPLE_DESCRIPTION_TEXT = L"This sample demonstrates secure communication xbox and custom web services.";
const HelpButtonAssignment HELP_BUTTONS[] = {
    { HelpID::A_BUTTON, L"Send a request to the Xbox people service" },
    { HelpID::X_BUTTON, L"Send a request to the first custom url" },
    { HelpID::Y_BUTTON, L"Send a request to the second custom url" },
    { HelpID::B_BUTTON, L"Cancel the current request" },
    { HelpID::BACK_BUTTON, L"Show/Hide Help" },
    { HelpID::LEFT_STICK, L"(Click) Take Screenshot" }
};

// List of application UI states
enum APPSTATE
{
    APP_INIT,
    APP_SIGNIN,
    APP_DONE_ERROR,
    APP_DONE,
};

//--------------------------------------------------------------------------------------
// Name: Sample
// Desc: Global class for this sample, derived from SampleFramework and
//       implementing Initialize(), Update() and Render().
//--------------------------------------------------------------------------------------

class Sample : public SampleFramework
{
public:

    virtual void Initialize() override;
    virtual void Update( float timeTotal, float timeDelta ) override;
    virtual void Render() override;

private:   

    HRESULT GetSignedInUserInfo();
    void    CallWebService( const std::wstring& url, bool bUseXSTSToken );
    void    ReadServiceResponse();

    XSF::Help       m_help;
    BitmapFont      m_font;
    Input           m_input;

    // Current state of the app.
    APPSTATE        m_appState;

    INT             m_iUserIndex;
    std::wstring    m_wstrUserGamerTag;
    std::wstring    m_wstrUserXuid;

    std::shared_ptr<HttpRequest>   m_pRequest;
    HRESULT         m_hrRequestResult;
    DWORD           m_dwRequestHTTPStatus;

    std::wstring    m_wstrRequestMethod;
    std::wstring    m_wstrRequestBody;

    DWORD           m_dwTimeoutInSec;
    BasicTimer      m_timer;

    // A copy of the HTTP results for display.
    std::wstring    m_wstrRequestURL;
    std::wstring    m_wstrResponseHeaders;
    std::wstring    m_wstrResponseData;
};

//--------------------------------------------------------------------------------------
// Name: Sample::Initialize()
// Desc: Initialize state before running the sample.
//--------------------------------------------------------------------------------------

void Sample::Initialize()
{
    // Get the D3DDevice.
    XSF::D3DDevice* const pDev = GetDevice();

    // Initialize the help screen
    const Settings& settings = SampleSettings();
    m_help.Initialize( pDev, settings.m_frameBufferWidth, settings.m_frameBufferHeight );
    m_help.SetHelpText( SAMPLE_TITLE_TEXT, SAMPLE_DESCRIPTION_TEXT, HELP_BUTTONS, ARRAYSIZE( HELP_BUTTONS ) );

    // Setup the UI states.
    XSF_ERROR_IF_FAILED( m_input.Initialize() );
    XSF_ERROR_IF_FAILED( m_font.Create( pDev, L"Media\\Fonts\\Arial_16" ) );

    // Start in initialization mode.
    m_appState = APP_INIT;

    m_iUserIndex = 0;

    m_pRequest = nullptr;
    m_hrRequestResult = S_OK;
    m_dwRequestHTTPStatus = 0;

    m_wstrRequestMethod = L"GET";
    m_dwTimeoutInSec = 10;
}

//--------------------------------------------------------------------------------------
// Name: Sample::CallWebService()
// Desc: Construct the HTTP request to the given URL
//--------------------------------------------------------------------------------------
void Sample::CallWebService( const std::wstring& ServiceURL, bool bUseXSTSToken  )
{
    XSF_ASSERT( !ServiceURL.empty() );

    if( m_pRequest )
    {
        return;
    }

    m_hrRequestResult = S_OK;
    m_dwRequestHTTPStatus = 0;
    m_wstrResponseHeaders.clear();
    m_wstrResponseData.clear();

    XSFScopedNamedEvent( nullptr, 0, L"Sample::CallWebService" );

    m_pRequest = std::make_shared<HttpRequest>();
    if( !m_pRequest )
    {
        m_hrRequestResult = E_OUTOFMEMORY;
        return;
    }

    // The HttpRequest object in the sample framework provides a mechanism to fetch the XSTS token and
    // signature for the calling user. However this implementation is synchronous and fetches the token
    // for every request and is only provided for demonstrative purposes. 
    // In a real world scenario, you would ideally use automatic token insertion by setting up your 
    // webservice endpoints and NSAL definitions on the XDP (Xbox Development Portal). 
    // If you are unable to, or choose to not take that route then you should cache the token and 
    // signature retreived by GetTokenAndSignatureAsync for your service so that you don't have
    // to fetch it before every call to your service. You can use this token until it's valid and
    // and renew it by calling GetTokenAndSignatureAsync again.
    m_pRequest->SetUseXSTSToken( bUseXSTSToken );

    m_pRequest->SetTimeout( m_dwTimeoutInSec*1000 );

    std::wstring wstrContentType;
    const BYTE* contentBytes = nullptr;
    size_t contentLength = 0;
    if( m_wstrRequestMethod == L"POST" )
    {
        contentBytes = reinterpret_cast<const BYTE*>(m_wstrRequestBody.data());
        contentLength = m_wstrRequestBody.size() * sizeof(wchar_t);
        wstrContentType = L"application/json\r\n";
    }

    // NOTE: See Open() in HttpRequest.cpp for how to get the Token, Signature, and add them to the request
    HRESULT hr = m_pRequest->Open( User::Users->GetAt(m_iUserIndex), m_wstrRequestMethod, ServiceURL, wstrContentType, contentBytes, contentLength );
    if( FAILED (hr) )
    {
        m_pRequest.reset();
        m_hrRequestResult = hr;
        m_appState = APP_DONE_ERROR;
        return;
    }

    XSF::DebugPrint("Request Started at %0.2f\n", m_timer.Update());
    m_appState = APP_DONE;

    return;
}

//--------------------------------------------------------------------------------------
// Name: Sample::ReadServiceResponse()
// Desc: Check the status of the request and copy the relevant information for the UI
//--------------------------------------------------------------------------------------
void Sample::ReadServiceResponse()
{
    if( m_pRequest && m_pRequest->IsResponseReceived() )
    {
        XSFScopedNamedEvent( nullptr, 0, L"Sample::ReadServiceResponse" );

        // Get results
        m_hrRequestResult = m_pRequest->GetResult();
        m_dwRequestHTTPStatus = m_pRequest->GetStatus();

        XSF::DebugPrint( "Request Completed at %0.2f, HResult:0x%08x Status:%d n", 
            m_timer.Update(), m_hrRequestResult, m_dwRequestHTTPStatus );
        
		// Warning: This function call increases the attack surface of your application. Be sure to review your data flows 
		// to ensure that they cannot be modified by an attacker. For more information, see http://microsoft.com/sdl

        m_wstrResponseHeaders = m_pRequest->GetHeaders();
        if( SUCCEEDED(m_hrRequestResult) )
        {
            const auto& buffer = m_pRequest->GetDataBuffer();
            if (buffer.size() < 64*1024)
            {
                // Attempt to stringify result for display if it's less than 64k
                m_wstrResponseData = m_pRequest->GetDataAsUTF8String();
            
                // Wrap the returned data text to fit output window
                for( int i = 100; i < m_wstrResponseData.size(); i += 100 )
                {
                    m_wstrResponseData.insert( i, L"\n" );
                }
            }
            else
            {
                WCHAR text[256];
                swprintf_s( text, L"%zd bytes", buffer.size() );
                m_wstrResponseData = text;
            }
        }

        if( FAILED(m_hrRequestResult) || 200 != m_dwRequestHTTPStatus )
        {
            m_appState = APP_DONE_ERROR;
        }
        else
        {
            m_appState = APP_DONE;
        }

        // this request is done, delete it.
        m_pRequest.reset();
    }
}

//--------------------------------------------------------------------------------------
// Name: Sample::Update()
// Desc: Called once per frame to update data
//--------------------------------------------------------------------------------------
void Sample::Update( float /*timeTotal*/, float timeDelta )
{
    // Get gamepad state
    const XSF::GamepadReading& input = GetGamepadReading();

    // Help screen grabs input exclusively if it's active.
    if( m_help.Update( input, timeDelta ) )
    {
        return;
    }

    // No Help screen active, app gets the input.
    switch( m_appState )
    {

    case APP_INIT:
        {
            // Do nothing.
            m_wstrResponseHeaders.clear();
            m_wstrResponseData.clear();
            m_wstrUserGamerTag.clear();
            m_wstrUserXuid.clear();
            m_hrRequestResult = 0;
            m_dwRequestHTTPStatus = 0;

            m_wstrRequestBody = CUSTOMURI_REQUEST_BODY;

            // Check for a signed-in user and get their info
            GetSignedInUserInfo();

            m_appState = APP_DONE;
        }
        break;

    case APP_SIGNIN:
        __fallthrough;
    case APP_DONE_ERROR:
        __fallthrough;
    case APP_DONE:
        {
            if( m_pRequest )
            {
                ReadServiceResponse();
            }
            
            if( input.WasAnythingPressedOrReleased() )
            {
                if( S_OK == GetSignedInUserInfo() )
                {
#if ENABLE_DELEGATED_URLS
                    
                    {
                        //  Friends call
                        if( input.APreviouslyPressed() )
                        {
                            m_wstrRequestMethod = L"GET";
                            m_wstrRequestURL = DELEGATED_REQUEST_FRIENDS;
                            CallWebService( m_wstrRequestURL, CUSTOMURI_REQUIRE_XSTS );
                        }

                        //  Inventory call
                        if( input.YPreviouslyPressed() )
                        {
                            m_wstrRequestMethod = L"GET";
                            m_wstrRequestURL = DELEGATED_REQUEST_INVENTORY;
                            CallWebService( m_wstrRequestURL, CUSTOMURI_REQUIRE_XSTS );
                        }
                        
                        //  Consume call
                        if( input.XPreviouslyPressed() )
                        {
                            m_wstrRequestMethod = L"POST";
                            m_wstrRequestURL = DELEGATED_REQUEST_CONSUME;
                            m_wstrRequestBody = L"{\"productId\":\"" CONSUMABLE_PRODUCT_ID L"\", \"quantity\": 1 }";
                            CallWebService( m_wstrRequestURL, CUSTOMURI_REQUIRE_XSTS );
                        }

                        //  Purchase call
                        if( input.BPreviouslyPressed() )
                        {
                            m_wstrRequestMethod = L"POST";
                            m_wstrRequestURL = DELEGATED_REQUEST_PURCHASE;

                            //Get the console's current region code / MarketplaceID (example: en-us, en-ca, fr-ca, etc.
                            std::wstring region = Windows::Xbox::ApplicationModel::Store::Configuration::MarketplaceId->Data();
                            m_wstrRequestBody = L"{\"region\":\"" + region + L"\"}";

                            CallWebService( m_wstrRequestURL, true );
                        }                      
                    }
#else
                    {
                        //
                        //  Xbox web service call
                        //
                        if( input.APreviouslyPressed() )
                        {
                            // People service to view a user's friends
                            m_wstrRequestURL = L"https://social.xboxlive.com/users/xuid(" + m_wstrUserXuid + L")/people";
                            CallWebService( m_wstrRequestURL, true );
                        }

                        //
                        //  Custom web service calls
                        //
                        if( input.XPreviouslyPressed() )
                        {
                            m_wstrRequestURL = CUSTOMURI_REQUEST_URL_1;
                            CallWebService( m_wstrRequestURL, CUSTOMURI_REQUIRE_XSTS );
                        }

                        if( input.YPreviouslyPressed() )
                        {
                            m_wstrRequestURL = CUSTOMURI_REQUEST_URL_2;
                            CallWebService( m_wstrRequestURL, CUSTOMURI_REQUIRE_XSTS );
                        }

                        // Check for cancellation
                        if( input.BPreviouslyPressed() )
                        {
                            if( m_pRequest != nullptr )
                            {
                                m_pRequest->Abort();

                                // We can choose to destroy the request object immediately
                                // OR we can for Abort to trigger completion and destroy 
                                // via the normal path in ReadServiceResponse
                                m_pRequest.reset();
                                m_hrRequestResult = E_ABORT;
                            }
                        }

                        if( input.DPadLeftPreviouslyPressed() )
                        {
                            if( m_dwTimeoutInSec > 0 )
                                m_dwTimeoutInSec--;
                        }
                        else if( input.DPadRightPreviouslyPressed() )
                        {
                            // Arbitrarily capped at 5 minutes for the sample
                            if( m_dwTimeoutInSec < 5*60 )
                                m_dwTimeoutInSec++;
                        }

                        if( input.DPadUpPreviouslyPressed() || input.DPadDownPreviouslyPressed() )
                        {
                            // Alternate between GET and POST for the sample
                            if( m_wstrRequestMethod == L"GET" )
                            {
                                m_wstrRequestMethod = L"POST";
                            }
                            else
                            {
                                m_wstrRequestMethod = L"GET";
                            }
                        }
                    }
#endif
                }
            }
        }
        break;
    }
}

//--------------------------------------------------------------------------------------
// Name: Sample::GetSignedInUserInfo()
// Desc: Checks for a signed in user and gets that user's info to display in the UI.
//       If no user is signed-in, app state is set to wait for a signed in user
//--------------------------------------------------------------------------------------
HRESULT Sample::GetSignedInUserInfo()
{
    HRESULT hr = E_FAIL;

    if( User::Users->Size > 0 )
    {
        m_appState = APP_DONE;
        auto user = User::Users->GetAt( m_iUserIndex );
        m_wstrUserXuid = user->XboxUserId->Data();
        m_wstrUserGamerTag = user->DisplayInfo->Gamertag->Data();
        hr = S_OK;
    }
    else
    {
        // No user, so clear related data
        m_appState = APP_SIGNIN;
        m_wstrUserXuid.clear();
        m_wstrUserGamerTag.clear();
        hr = E_PENDING;
    }

    return hr;
}

//--------------------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame to render the output text
//--------------------------------------------------------------------------------------
void Sample::Render()
{
    const DWORD TITLE_COLOR   = 0xFFFFFF00;
    const DWORD TEXT_COLOR    = 0xFFFFFFFF;
    const DWORD RESULT_COLOR  = 0xFFA0FFA0;
    const DWORD ERROR_COLOR   = 0xFFAA3333;

    XSF::D3DDeviceContext* const pCtx = GetImmediateContext();

    // Help screen take exclusive control of the screen when it's active
    if( m_help.IsVisible() )
    {
        m_help.Render( pCtx );
        return;
    }

    // We divide the screen into a course grid of 16x10 cells to make placing UI
    // elements easier
    const D3D11_RECT screenRect = m_font.GetWindow();
    const FLOAT fWidth  = static_cast<FLOAT>( screenRect.right  - screenRect.left ) - 1.0f;
    const FLOAT fHeight = static_cast<FLOAT>( screenRect.bottom - screenRect.top )  - 1.0f;
    const FLOAT fGridX = fWidth / 16.0f;
    const FLOAT fGridY = fHeight / 10.0f;
    
    // temporary wchar buffer for fomatted strings.
    const INT iTextBufferMax = 1023;
    WCHAR text[iTextBufferMax+1];

    // Draw the text
    //
    // NOTE: do this after other draws as setting state inside a font.Begin()/font.End()
    // block can mess up the font draw states
    m_font.Begin( pCtx );

    // Draw the title
    m_font.SetScaleFactors( 1.5f, 1.5f );
    m_font.DrawText( 1 * fGridX,
                     1 * fGridY,
                     TITLE_COLOR,
                     L"Web Services Sample" );

    // Draw instructions and user information
    m_font.SetScaleFactors( 1.2f, 1.2f );
#if ENABLE_DELEGATED_URLS
    {
        swprintf_s( text, 
                    iTextBufferMax,
                    L"Press " GLYPH_A_BUTTON L" to call " DELEGATED_REQUEST_FRIENDS   L"\n"
                    L"Press " GLYPH_X_BUTTON L" to call " DELEGATED_REQUEST_CONSUME   L"\n"
                    L"Press " GLYPH_Y_BUTTON L" to call " DELEGATED_REQUEST_INVENTORY L"\n"
                    L"Press " GLYPH_B_BUTTON L" to call " DELEGATED_REQUEST_PURCHASE  L"\n" );
    }
#else
    {
        swprintf_s( text, 
                    iTextBufferMax,
                    L"Press " GLYPH_A_BUTTON L" to call Xbox people service\n"
                    L"Press " GLYPH_X_BUTTON L" to call " CUSTOMURI_REQUEST_URL_1 L"\n"
                    L"Press " GLYPH_Y_BUTTON L" to call " CUSTOMURI_REQUEST_URL_2 L"\n"
                    L"Press " GLYPH_B_BUTTON L" to cancel request\n"
                    L"Press " GLYPH_UD_ARROW L" to change the method, Current: %s\n"
                    L"Press " GLYPH_LR_ARROW L" to change the timeout, Current: %d seconds\n",
                    m_wstrRequestMethod.c_str(),
                    m_dwTimeoutInSec );
    }
#endif
    m_font.DrawText( 1 * fGridX,
                     2 * fGridY,
                     TEXT_COLOR,
                     text );

    if( m_wstrUserXuid != L"" )
    {
        swprintf_s( text,
                    iTextBufferMax,
                    L"User: %s",
                    m_wstrUserGamerTag.c_str() );
    }
    else
    {
        swprintf_s( text,
                    iTextBufferMax,
                    L"User: No profile loaded, sign-in and select service to call");
    }
    m_font.DrawText( 10 * fGridX,
                     1 * fGridY,
                     m_wstrUserXuid == L"" ? ERROR_COLOR : TEXT_COLOR,
                     text );

    if( m_appState == APP_DONE || m_appState == APP_DONE_ERROR )
    {
        if( m_pRequest )
        {
            m_font.DrawText( 10 * fGridX,
                             2 * fGridY,
                             TITLE_COLOR,
                             L"Waiting for Response..." );
        }
        else if( m_dwRequestHTTPStatus != 0 || m_hrRequestResult != S_OK )
        {
            // Draw the request's result and status
            swprintf_s( text,
                        iTextBufferMax,
                        L"HRESULT:      0x%08x\nHTTP Status: %u", 
                        m_hrRequestResult, m_dwRequestHTTPStatus );
            m_font.DrawText( 10 * fGridX,
                             2 * fGridY,
                             m_appState == APP_DONE_ERROR ? ERROR_COLOR : TEXT_COLOR, 
                             text );
        }

        m_font.SetScaleFactors( 1.0f, 1.0f );
        if( !m_wstrRequestURL.empty() )
        {
            // Draw the request URL
            swprintf_s( text,
                        iTextBufferMax,
                        L"Request URL:\n%s", 
                        m_wstrRequestURL.c_str() );
            m_font.DrawText( 1 * fGridX,
                             4 * fGridY,
                             TITLE_COLOR, 
                             text );
        }

        if( !m_wstrResponseHeaders.empty() || !m_wstrResponseData.empty() )
        {
            // Draw the response headers and data 
            const INT maxHeaderLength = 512;
            std::wstring headerText = m_wstrResponseHeaders;
            if( headerText.size() > maxHeaderLength )
            {
                headerText = m_wstrResponseHeaders.substr(0, maxHeaderLength);
            }

            _snwprintf_s( text, 
                          iTextBufferMax, 
                          L"Response Headers:\n%sResponse Data:\n%s", 
                          headerText.c_str(), m_wstrResponseData.c_str() );
            m_font.DrawText( 1 * fGridX,
                             5 * fGridY,
                             RESULT_COLOR,
                             text );
        }
    }

    // Done with font rendering, reset the graphics context.
    m_font.End();
}

XSF_DECLARE_ENTRY_POINT( Sample );
